require_relative "base_helper"

describe Toolchain::Attributes, "Attributes (Method)" do

  let(:model) { ModelData.new }

  def expect_date_time
    [
      Date.today.tap { |d| Date.stubs(:today).returns(d) },
      Time.now.tap { |t| Time.stubs(:now).returns(t) },
      DateTime.now.tap { |dt| DateTime.stubs(:now).returns(dt) }
    ]
  end

  it "should return attributes with Symbol keys" do
    date, time, date_time = expect_date_time

    model.attributes.should == {
      array_default: [],
      boolean_default: false,
      float_default: 0.0,
      hash_default: {},
      integer_default: 0,
      string_default: "",
      date_default: date,
      time_default: time,
      date_time_default: date_time
    }
  end

  it "should return attributes with String keys" do
    Toolchain::Attributes::Configuration.
      hash_transformation = :stringify_keys

    date, time, date_time = expect_date_time

    model.attributes.should == {
      "array_default" => [],
      "boolean_default" => false,
      "float_default" => 0.0,
      "hash_default" => {},
      "integer_default" => 0,
      "string_default" => "",
      "date_default" => date,
      "time_default" => time,
      "date_time_default" => date_time
    }
  end

  it "should symbolize hash, hash_default" do
    model.hash = { "key" => { "value" => true } }
    model.hash_default = { "key" => { "value" => true } }
    model.attributes.tap do |attrs|
      attrs[:hash].should == { key: { value: true } }
      attrs[:hash_default].should == { key: { value: true } }
    end
  end

  it "should stringify hash, hash_default" do
    Toolchain::Attributes::Configuration.
      hash_transformation = :stringify_keys

    model.hash = { key: { value: true } }
    model.hash_default = { key: { value: true } }
    model.attributes.tap do |attrs|
      attrs["hash"].should == { "key" => { "value" => true } }
      attrs["hash_default"].should == { "key" => { "value" => true } }
    end
  end

  it "should include nil attributes" do
    Toolchain::Attributes::Configuration.
      include_nil_in_attributes = true

    date, time, date_time = expect_date_time

    model.attributes.should == {
      array: nil,
      array_default: [],
      boolean: nil,
      boolean_default: false,
      float: nil,
      float_default: 0.0,
      hash: nil,
      hash_default: {},
      integer: nil,
      integer_default: 0,
      string: nil,
      string_default: "",
      date: nil,
      date_default: date,
      time: nil,
      time_default: time,
      date_time: nil,
      date_time_default: date_time
    }
  end

  it "should mass-assign attributes" do
    date, time, date_time = expect_date_time

    model.attributes = {
      string: "Example",
      integer: 123,
      float: 1.23,
      hash: { foo: "bar" },
      array: [1, 2, 3],
      boolean: true,
      date: date,
      date_default: date,
      time: time,
      time_default: time,
      date_time: date_time,
      date_time_default: date_time,
    }

    model.attributes.should == {
      array: [1, 2, 3],
      array_default: [],
      boolean: true,
      boolean_default: false,
      float: 1.23,
      float_default: 0.0,
      hash: { foo: "bar" },
      hash_default: {},
      integer: 123,
      integer_default: 0,
      string: "Example",
      string_default: "",
      date: date,
      date_default: date,
      time: time,
      time_default: time,
      date_time: date_time,
      date_time_default: date_time
    }
  end

  it "should not assign undefined attributes" do
    date, time, date_time = expect_date_time
    model.attributes = { unknown_attribute: true }
    model.attributes.should == {
      array_default: [],
      boolean_default: false,
      float_default: 0.0,
      hash_default: {},
      integer_default: 0,
      string_default: "",
      date_default: date,
      time_default: time,
      date_time_default: date_time,
    }
  end

  it "should raise an exception non-Hash type is passed" do
    expect { model.attributes = 0 }.
      to raise_error(
        Toolchain::Attributes::Errors::InvalidMassAssignment,
        "Can't mass-assign Fixnum (0) type to ModelData#attributes.")
  end

  class InheritedModelData < ModelData
  end

  it "should inherit all attribute definitions" do
    date, time, date_time = expect_date_time
    InheritedModelData.new.attributes.should == {
      array_default: [],
      boolean_default: false,
      float_default: 0.0,
      hash_default: {},
      integer_default: 0,
      string_default: "",
      date_default: date,
      time_default: time,
      date_time_default: date_time
    }
  end

  it "should mass-assign superclass keys" do
    date, time, date_time = expect_date_time
    model = InheritedModelData.new
    model.attributes = { string: "Michael" }
    model.string.should == "Michael"
  end

  it "should mass-assing superclass string keys" do
    date, time, date_time = expect_date_time
    model = InheritedModelData.new
    model.attributes = { "string" => "Michael" }
    model.string.should == "Michael"
  end
end
